using System;
using System.Collections.Generic;
using System.Linq;

namespace Graph.Core
{
    public class EulerMethodStructure<TVertex, TEdge> : GraphStructure<TVertex, TEdge>
        where TVertex : EulerVertex, IStructOwner<TVertex>
        where TEdge : IPhysik2DEdge<TVertex>
    {
        private List<Physical2DGraphVertexCluster> clusters = new List<Physical2DGraphVertexCluster>();

        public EulerMethodStructure(IGraph<TVertex, TEdge> graph)
            : base(graph)
        {
        }

        public List<Physical2DGraphVertexCluster> Clusters
        {
            get { return clusters; }
        }

        private double difference;
        private double difference_priorStep;
        private double difference_priorPriorStep;

        protected override void PreSteping(NumericalMethodParametres parametres)
        {
            if (parametres.DontCalculateRemoteNodesGravitation)
            {
                DivideByCluster(graph.Vertices.ToList(), parametres.NearestNodesCountThreshold);

                foreach (var cluster in clusters)
                    cluster.Position = cluster.Vertices.CalcMassCenter();

                foreach (var vertex in graph.Vertices)
                    FillNearesNodes(vertex, parametres.NearestNodesCountThreshold);
            }
        }

        public void FillNearesNodes(TVertex vertex, int count)
        {
            if (vertex.ClusterID != -1)
            {
                vertex.OtherNodesFromCluster = Clusters[vertex.ClusterID].Vertices.ToList();
                vertex.OtherNodesFromCluster.Remove(vertex);
            }
            else
                vertex.OtherNodesFromCluster.Clear();

            vertex._remoteNodesGravitationForce = vertex.OtherNodesFromCluster.CalcGravitationForceByAll(Environment, vertex);

            #region OLD

            //var nearestNodes = OwnerGraph.Vertices.ToList();
            //nearestNodes.Remove(this);

            //nearestNodes.Sort(DistanceComparison); // heavy operation

            //_remoteNodesGravitationForce.X = 0;
            //_remoteNodesGravitationForce.Y = 0;

            //if (nearestNodes.Count > count)
            //    nearestNodes.RemoveRange(count, nearestNodes.Count - count);

            //foreach (var vertex in nearestNodes)
            //{
            //    var direction = Position - vertex.Position;
            //    double distanceSquared = direction.LengthSquared;
            //    double GFToSingleNode_Value = OwnerGraph.PhysicalModelParametres.NodeGravitationCoefficient / distanceSquared;
            //    var vGFToSingleNode = direction * GFToSingleNode_Value;
            //    _remoteNodesGravitationForce = _remoteNodesGravitationForce + vGFToSingleNode;
            //}

            #endregion
        }

        protected override void ProcessOneStep(IList<TVertex> vertices, NumericalMethodParametres parametres)
        {
            ProcessOneStep_Euler(vertices, parametres);
            //ProcessOneStep_Potencial_LocalNewton(vertices, parametres as NewtonNumericalMethodParametres);
        }

        protected void ProcessOneStep_Euler(IList<TVertex> vertices, NumericalMethodParametres parametres)
        {
            double timeInterval = parametres.OneIterationTimeStep;

            if (parametres.AutoDecreaseTimeStep)
                if (difference > 0)
                    timeInterval = Math.Min(timeInterval, 0.2D / (Math.Max(Math.Max(difference, difference_priorStep), difference_priorPriorStep)));

            difference_priorPriorStep = difference_priorStep;
            difference_priorStep = difference;

            foreach (var vertex in vertices)
            {
                var gravForce = CalcGravitationForce(vertices, vertex);
                var stiffForce = vertex.CalcStiffnessForce(graph.AdjacentEdges);
                var dissForce = vertex.CalcDissipativeForce(Environment);

                vertex.force = gravForce + stiffForce + dissForce;

                var vAcseleration = vertex.force * Environment.NodeInertialMass;
                vertex._newVelocity = vertex.velocity + (vAcseleration * timeInterval);
                vertex._newPosition = vertex.Position + (vertex._newVelocity * timeInterval);
            }

            difference = vertices.CalcDifference();
            UpdateNewProperties(vertices);
        }

        protected void ProcessOneStep_Potencial_LocalNewton(IList<TVertex> Vertices, NewtonNumericalMethodParametres parametres)
        {
            //    1.          
            //    2.   
            //    3.    }
            //       )    ()
            //       )    ()

            //   ,   )
            if (parametres.IsMoveNow)
            {
                difference = 0;
            }


            foreach (var vertex in Vertices)
            {
                //  --- 1.      
                vertex.force = CalcForce(Vertices, vertex);

                double dx = 0;
                double dy = 0;
                if (parametres.isTwoCoords)
                {
                    // 2.   
                    Vector Fdx = CalcForceDerivative(Vertices, vertex, new Vector(1, 0), parametres, vertex.force);
                    Vector Fdy = CalcForceDerivative(Vertices, vertex, new Vector(0, 1), parametres, vertex.force);
                    // 3.    .   
                    //
                    //    F+A(dP)=0,
                    //
                    //    dp = -F*A^(-1);
                    //  dP = [dx, dy]^T; ()
                    //     A = [[Fdx.X, Fdy.X], [Fdx.Y, Fdy.Y]]
                    //      
                    // [a  b]-1     1    [ d  -b]
                    //          = -----  
                    // [c  d]     ad-bc  [-c   a]
                    double detA = Fdx.X * Fdy.Y - Fdy.X * Fdx.Y;
                    if (Math.Abs(detA - 0) > double.Epsilon)
                    {
                        dx = -(vertex.force.X * Fdy.Y - vertex.force.Y * Fdy.X) / detA;
                        dy = -(-vertex.force.X * Fdx.Y + vertex.force.Y * Fdx.X) / detA;
                    }
                }
                else
                {
                    //    -    .
                    // 2.   
                    Vector Fdx = CalcForceDerivative(Vertices, vertex, new Vector(1, 0), parametres, vertex.force);
                    // 3.    .
                    dx = -vertex.force.X / Fdx.X;
                    dx = dx * parametres.BrakeFactor;


                    Vector Fdy = CalcForceDerivative(Vertices, vertex, new Vector(0, 1), parametres, vertex.force);
                    dy = -vertex.force.Y / Fdy.Y;
                    dy = dy * parametres.BrakeFactor;
                }

                //    
                double s2 = dx * dx + dy * dy;
                double d = Math.Sqrt(s2);

                if (parametres.MaxShift < d)
                {
                    dx = dx * parametres.MaxShift / d;
                    dy = dy * parametres.MaxShift / d;
                    d = parametres.MaxShift; // , ..    
                }

                vertex._newPosition.X = vertex.Position.X + dx;
                vertex._newPosition.Y = vertex.Position.Y + dy;

                if (parametres.IsMoveNow)
                {
                    // )    ()
                    //    .
                    difference += Math.Sqrt(d);

                    vertex.Position = vertex._newPosition;
                }
                // )    ()
            }

            if (parametres.IsMoveNow)
            {
                //    )
                difference = difference / Vertices.Count();
            }
            else
            {
                //    a)
                difference = Vertices.CalcDifference();
                UpdateNewProperties(Vertices);
            }
        }

        protected void ProcessOneStep_RungeKutta4(IList<TVertex> Vertices, NumericalMethodParametres parametres)
        {
            double dt = parametres.OneIterationTimeStep;
            //  
            foreach (var vertex in Vertices)
            {
                //   
                vertex._oldPosition = vertex.Position;
                vertex._oldVelocity = vertex.velocity;
                //   
                Vector F = CalcForce(Vertices, vertex);
                F = Vector.Add(F, vertex.CalcDissipativeForce(Environment));

                //    
                vertex._xk1 = vertex.velocity.X * dt;
                vertex._xm1 = F.X * dt;

                vertex._yk1 = vertex.velocity.Y * dt;
                vertex._ym1 = F.Y * dt;
            }

            //     
            foreach (var vertex in Vertices)
            {
                vertex.Position.X = vertex._oldPosition.X + vertex._xk1 / 2;
                vertex.velocity.X = vertex._oldVelocity.X + vertex._xm1 / 2;

                vertex.Position.Y = vertex._oldPosition.Y + vertex._yk1 / 2;
                vertex.velocity.Y = vertex._oldVelocity.Y + vertex._ym1 / 2;
            }

            //  
            foreach (var vertex in Vertices)
            {
                Vector F = CalcForce(Vertices, vertex);
                F = Vector.Add(F, vertex.CalcDissipativeForce(Environment));
                //   
                vertex._xk2 = vertex.velocity.X * dt;
                vertex._xm2 = F.X * dt;

                vertex._yk2 = vertex.velocity.Y * dt;
                vertex._ym2 = F.Y * dt;
            }

            //     
            foreach (var vertex in Vertices)
            {
                vertex.Position.X = vertex._oldPosition.X + vertex._xk2 / 2;
                vertex.velocity.X = vertex._oldVelocity.X + vertex._xm2 / 2;

                vertex.Position.Y = vertex._oldPosition.Y + vertex._yk2 / 2;
                vertex.velocity.Y = vertex._oldVelocity.Y + vertex._ym2 / 2;
            }

            //  
            foreach (var vertex in Vertices)
            {
                Vector F = CalcForce(Vertices, vertex);
                F = Vector.Add(F, vertex.CalcDissipativeForce(Environment));

                //   
                vertex._xk3 = vertex.velocity.X * dt;
                vertex._xm3 = F.X * dt;

                vertex._yk3 = vertex.velocity.Y * dt;
                vertex._ym3 = F.Y * dt;
            }

            //     
            foreach (var vertex in Vertices)
            {
                vertex.Position.X = vertex._oldPosition.X + vertex._xk3;
                vertex.velocity.X = vertex._oldVelocity.X + vertex._xm3;

                vertex.Position.Y = vertex._oldPosition.Y + vertex._yk3;
                vertex.velocity.Y = vertex._oldVelocity.Y + vertex._ym3;
            }

            //  
            foreach (var vertex in Vertices)
            {
                Vector F = CalcForce(Vertices, vertex);
                F = Vector.Add(F, vertex.CalcDissipativeForce(Environment));
                //   
                vertex._xk4 = vertex.velocity.X * dt;
                vertex._xm4 = F.X * dt;

                vertex._yk4 = vertex.velocity.Y * dt;
                vertex._ym4 = F.Y * dt;
            }

            //   +    (   )
            // ...
            //      _position,     _newPosition
            //    CalcDifference   (  )
            // (   UpdateNewProperties)
            foreach (var vertex in Vertices)
            {
                vertex._newPosition.X = vertex._oldPosition.X +
                                        (vertex._xk1 + 2 * vertex._xk2 + 2 * vertex._xk3 + vertex._xk4) / 6;
                vertex._newPosition.Y = vertex._oldPosition.Y +
                                        (vertex._yk1 + 2 * vertex._yk2 + 2 * vertex._yk3 + vertex._yk4) / 6;
                vertex._newVelocity.X = vertex._oldVelocity.X +
                                        (vertex._xm1 + 2 * vertex._xm2 + 2 * vertex._xm3 + vertex._xm4) / 6;
                vertex._newVelocity.Y = vertex._oldVelocity.Y +
                                        (vertex._ym1 + 2 * vertex._ym2 + 2 * vertex._ym3 + vertex._ym4) / 6;
                vertex.Position = vertex._oldPosition;
                vertex.velocity = vertex._oldVelocity;
            }

            //!//difference = Vertices.CalcDifference();
            UpdateNewProperties(Vertices);

        }

        #region    (, , , )

        private Vector CalcForce(IEnumerable<TVertex> Vertices, TVertex vertex)
        {
            return CalcGravitationForce(Vertices, vertex) + vertex.CalcStiffnessForce(GetEdges);
        }

        private Vector CalcGravitationForceByClusters(TVertex vertex)
        {
            return vertex.OtherNodesFromCluster.CalcGravitationForceByAll(Environment, vertex) +
                   vertex._remoteNodesGravitationForce;
        }

        private Vector CalcGravitationForce(IEnumerable<TVertex> vertices, TVertex vertex)
        {
            //return vertex.CalcGravitationForceWithConnectedNodesGravitationInteractionOnly(Environment, graph.AdjacentEdges);

            return CalcGravitationForceByClusters(vertex);

            //return vertices.CalcGravitationForceByAll(Environment, vertex);
        }

        private Vector CalcForceDerivative(IEnumerable<TVertex> Vertices, TVertex vertex, Vector direction, NewtonNumericalMethodParametres p, Vector f0)
        {
            direction = direction / direction.Length;
            Vector p0 = vertex.Position; //    
            vertex.Position.X += p.Delta * direction.X; //       . .
            vertex.Position.Y += p.Delta * direction.Y; // 
            Vector f = CalcForce(Vertices, vertex); //    
            vertex.Position = p0; //  
            // derivative
            return (f - f0) / p.Delta;
        }

        #endregion

        private void UpdateNewProperties(IEnumerable<TVertex> Vertices)
        {
            foreach (var vertex in Vertices)
            {
                if (vertex._newVelocity.Length < Environment.MaxVelocity)
                    vertex.velocity = vertex._newVelocity;

                vertex._oldPosition = vertex.Position;
                vertex.Position = vertex._newPosition;
            }
        }

        public override void Reset()
        {
            base.Reset();
            difference = .0D;
            difference_priorStep = .0D;
            difference_priorPriorStep = .0D;
        }

        public void DivideByCluster(IList<TVertex> vertices, int nodesCountInCluster)
        {
            clusters.Clear();

            if (vertices.Count() == 0)
            {
                return;
            }

            foreach (var v in vertices)
            {
                v.ClusterID = -1;
            }

            //  
            int currentClusterID = 0;
            Physical2DGraphVertexCluster cluster = new Physical2DGraphVertexCluster();
            clusters.Add(cluster);
            cluster.ID = currentClusterID;


            //   
            var candidates = new Queue<TVertex>();

            //    

            var currentNode = vertices.First();
            currentNode.ClusterID = cluster.ID;
            cluster.Vertices.Add(currentNode);

            candidates.Enqueue(currentNode);

            //   nodesCountInCluster     ,     
            int foundCount = 1;

            //var verticles = vertices.OrderBy(i => Guid.NewGuid()).ToList();
            //foreach (var vertex in verticles)
            //{
            //    var v = vertex;
            //    if ((v.ClusterID == -1) && (foundCount < nodesCountInCluster))
            //    {
            //        //candidates.Enqueue(v);
            //        v.ClusterID = currentClusterID;
            //        cluster.Vertices.Add(v);
            //        foundCount++;
            //    }

            //    if (foundCount >= nodesCountInCluster)
            //    {
            //        currentClusterID++;
            //        foundCount = 0;
            //        cluster = new Physical2DGraphVertexCluster();
            //        cluster.ID = currentClusterID;
            //        clusters.Add(cluster);
            //    }
            //}

            while (candidates.Count > 0) //     
            {
                currentNode = candidates.Dequeue();

                foreach (var ed in GetEdges(currentNode))
                {
                    var v = ed.GetOtherVertex(currentNode);
                    if ((v.ClusterID == -1) && (foundCount < nodesCountInCluster))
                    {
                        candidates.Enqueue(v);
                        v.ClusterID = currentClusterID;
                        cluster.Vertices.Add(v);
                        foundCount++;
                    }

                    if (foundCount >= nodesCountInCluster)
                    {
                        currentClusterID++;
                        foundCount = 0;
                        cluster = new Physical2DGraphVertexCluster();
                        cluster.ID = currentClusterID;
                        clusters.Add(cluster);
                    }
                }
            }

            //      "" 
            if (cluster.Vertices.Count == 0)
            {
                clusters.Remove(cluster);
            }
        }
    }
}