using System;
using System.Collections.Generic;
using System.Linq;

namespace Graph.Core
{
    public abstract class GraphStructure<TVertex, TEdge> : IContainer<TVertex>
        where TEdge : IEdge<TVertex> 
        where TVertex : IPhysic2DVertex, IStructOwner<TVertex>
    {
        protected IGraph<TVertex, TEdge> graph;

        protected GraphStructure(IGraph<TVertex, TEdge> graph)
        {
            this.graph = graph;
            graph.VertexAdded += graph_VertexAdded;
            //Dictionary<int, List<int>> dic = new Dictionary<int, List<int>>();
            //var graph = dic.ToVertexAndEdgeListGraph(
            //    kv => Array.ConvertAll(kv.Value, v => new SEquatableEdge<int>(kv.Key, v))
            //    );
            NewVertexSeedParametres = new NewVertexSeedModel();
            Environment = new Environment();
            GetEdges = v => graph.AdjacentEdges(v);
        }

        public int EdgesCount(TVertex vertex)
        {
            return graph.AdjacentEdges(vertex).Count();
        }

        protected virtual void OnAddVertex(TVertex vertex)
        {
            (vertex as IStructOwner<TVertex>).OwnerGraph = this;
        }

        private void graph_VertexAdded(TVertex vertex)
        {
            OnAddVertex(vertex);
        }

        public Environment Environment { get; set; }

        private Vector massCenter;
        private double avgAngleVelocity;

        protected abstract void PreSteping(NumericalMethodParametres parametres);

        public void ProcessUnfoldingModel(NumericalMethodParametres parametres)
        {
            massCenter = graph.Vertices.CalcMassCenter();

            PreSteping(parametres);

            for (int iterationIndex = 0; iterationIndex < parametres.OneProcessIterationCount; iterationIndex++)
                ProcessOneStep(graph.Vertices.ToList(), parametres);

            //MoveMassCenterTo(graph.Vertices.ToList(), massCenter);

            //var av = graph.Vertices.CalcAvgRotationVelocity(massCenter);
            //Vertices.Rotate(massCenter, av); 
            //avgAngleVelocity = graph.Vertices.CalcAvgRotationVelocity(massCenter);
        }

        protected abstract void ProcessOneStep(IList<TVertex> vertices, NumericalMethodParametres p);

        public void MoveMassCenterTo(IList<TVertex> Vertices, Vector p)
        {
            var d = p - Vertices.CalcMassCenter();
            foreach (var vertex in Vertices)
            {
                vertex.Position.X += d.X;
                vertex.Position.Y += d.Y;
            }
        }

        public virtual void Reset()
        {
            massCenter = Vector.Empty;
        }

        public NewVertexSeedModel NewVertexSeedParametres { get; private set; }

        private void SetNewVertexPosition(TVertex vertex)
        {
            vertex.Position = NewVertexSeedParametres.GetNextPosition();
            
        }

        private void SetNewVertexPositionNearOldVertex(TVertex vertex, TVertex oldVertex)
        {
            vertex.Position = NewVertexSeedParametres.GetNextPositionNear(oldVertex.Position);
        }

        protected Func<TVertex, IEnumerable<TEdge>> GetEdges;
    }

    public class Environment
    {
        /// <summary>
        ///    
        /// </summary>  
        public double FreeConnectionLength = 20.0D;

        /// <summary>
        ///      
        /// </summary>
        public double NodeGravitationCoefficient = 200.0D;

        /// <summary>
        ///    (  )
        /// </summary>
        public double ConectionGripFactor = 0.5D;

        /// <summary>
        ///   =     
        /// </summary>
        public double NodeInertialMass = .1D;

        /// <summary>
        ///       .  -     (  )
        /// </summary>
        public bool IsWithConnectedNodesGravitationInteractionOnly;


        public Environment()
        {
        }

        public double ResistingForceFactor_0(Vector position)
        {
            return 1d;
        }

        public double ResistingForceFactor_1(Vector position)
        {
            return .5d;
        }

        public static double MaxVelocity { get; set; }
    }
}